module Postburner
  class Job < ApplicationRecord
    include Backburner::Queue

    LOG_LEVELS = [
      :debug,
      :info,
      :warning,
      :error
    ].freeze

    before_validation :ensure_sid!
    before_destroy :delete!

    validates :sid, presence: {strict: true}

    def queue!(options={})
      return if self.queued_at.present? && self.bkid.present?

      case
      when options[:at].present?
        # this is rudimentary, add error handling
        options[:delay] ||= options[:at].to_i - Time.zone.now.to_i
        update_column :run_at, options[:at]
        options.delete(:at)
      when options[:delay].present?
        update_column :run_at, Time.zone.now + options[:delay].seconds
      end

      insert!(options)
    end

    # tube: backburner.worker.queue.backburner-jobs
    #
    def self.perform(id, _={})
      begin
        job = self.find(id)
        job.perform!(job.args)
      rescue ActiveRecord::RecordNotFound => e
        Rails.logger.warn <<-MSG
[Postburner::Job] [#{id}] Not Found.
        MSG
      end
    end

    def perform!(args={})
      self.attempting

      self.update_columns(
        attempting_at:  self.attempting_at,
        attempts:       self.attempts,
        attempt_count:  self.attempts.length,
        lag:            self.lag,
        processing_at:  Time.zone.now,
      )

      begin
        if self.processed_at.present?
          self.log! "Already Processed", level: :error
          self.delete!
          return
        end

        if self.removed_at.present?
          self.log! "Removed", level: :error
          return
        end

        if self.run_at && self.run_at > Time.zone.now
          response = self.insert! delay: self.run_at - Time.zone.now
          self.log! "PREMATURE; RE-INSERTED: #{response}"
          return
        end

        self.log('START')

        self.perform(args)

        self.log('DONE')

        begin
          now = Time.zone.now
          _duration =  (now - self.processing_at) * 1000 rescue nil
          self.update_columns(
            processed_at: now,
            duration:     _duration,
            errata:       self.errata,
            error_count:  self.errata.length,
            logs:         self.logs,
            log_count:    self.logs.length,
          )
        rescue Exception => e
          self.log_exception!(e)
          self.log! '[Postburner] Could not set data after processing.'
          # TODO README doesn't retry if Postburner is to blame
        end

      rescue Exception => exception
        self.log_exception!(exception)
        self.log! '[Postburner] Exception raised during perform prevented completion.'
        raise exception
      end

    end

    def delete!
      return unless self.beanstalk_job
      begin
        self.beanstalk_job.delete
      rescue Beaneater::NotConnected => e
        self.beanstalk_job!.delete
      end
    end

    def kick!
      return unless self.beanstalk_job
      begin
        self.beanstalk_job.kick
      rescue Beaneater::NotConnected => e
        self.beanstalk_job!.kick
      end
    end

    def remove!
      if self.beanstalk_job
        self.delete!
        self.update_column(:removed_at, Time.zone.now)
      else
        false
      end
    end

    def beanstalk_job
      return unless self.bkid.present?
      return @_beanstalk_job if @_beanstalk_job

      @_beanstalk_job = Postburner.connection.beanstalk.jobs.find(self.bkid)

      @_beanstalk_job
    end

    def beanstalk_job!
      @_beanstalk_job = nil
      self.beanstalk_job
    end

    def log_exception(exception)
      self.errata << [
        Time.zone.now,
        {
          class:      exception.class,
          message:    exception.message,
          backtrace:  exception.backtrace,
        }
      ]
    end

    def log_exception!(exception)
      self.log_exception(exception)
      self.update_column :errata, self.errata
    end

    def log(message, options={})
      options[:level] ||= :info
      options[:level] = :error unless LOG_LEVELS.member?(options[:level])

      self.logs << [
        Time.zone.now,
        options[:level],
        message,
      ]
    end

    def log!(message, options={})
      self.log(message, options)
      self.update_column :logs, self.logs
    end

    private

    def insert!(options={})
      response = Backburner::Worker.enqueue(Postburner::Job, self.id, options)

      self.log("QUEUED: #{response}")

      update_columns(
        queued_at:  Time.zone.now,
        bkid:       response[:id],
        logs:       self.logs,
      )

      response
    end

    def attempting
      now = Time.zone.now
      self.attempts << now
      self.attempting_at ||= now
      self.lag ||= (self.attempting_at - self.intended_at) * 1000
      now
    end

    def intended_at
      self.run_at ? self.run_at : self.queued_at
    end

    def ensure_sid!
      self.sid ||= SecureRandom.uuid
    end

    def message
      "Job: #{self.id}"
    end

  end
end
